/*
 * dts.c:	Use the S/P-DIF interface for redirecting DTS stream (NO decoding!).
 *
 * Parts based on code by:
 *
 * Copyright (C) 2002 Unknown
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Remark: Note that DTS is NOT an open standard and is the property of
 *         the DIGITAL THEATER SYSTEMS Inc. and therefore (re-)distribution
 *         may violate this owner ship.
 *         ... Nevertheless, the DTS stream is forwarded from a DVB TV channel
 *         which has to pay its license fee and also the hardware decoder
 *         receiving the forwarded DTS stream was paid with this license fee.
 *
 * Copyright (C) 2003 Werner Fink, <werner@suse.de>
 */

#include "types.h"
#include "dts.h"

static frame_t & scan_dts_frame(const uint_8 *&out, const uint_8 *const tail);
#define DTS_BURST_SIZE	2048	// Not SPDIF_BURST_SIZE !
iec60958 dts(scan_dts_frame, 48000, DTS_BURST_SIZE);

typedef struct _syncinfo {
    uint_32	sampling_rate;	// Sampling rate in hertz (?)
    uint_16	bit_rate;	// Bit rate in kilobits (?)
    uint_16	frame_size;	// Frame size in 8 bit words
    uint_8	bsid;		// Bit stream identification
    uint_8	bsmod;		// Bit stream mode (?)
} syncinfo_t;

static const uint_32 magic = char2int(0x7f, 0xfe, 0x80, 0x01);

// This internal function determines from the first eight bytes after
// the four sync words the sampling rate(?), frame size, bit stream
// identification(?), and the bit stream mode(?).
static inline void parse_syncinfo(syncinfo_t &syncinfo, const uint_8 *data)
{
    switch((uint_8)((data[4] & 0x01) << 6) | ((data[5] >> 2) & 0x3f)) {
    case 0x0f:
	syncinfo.bsid = 0x0b;
	break;
    case 0x1f:
	syncinfo.bsid = 0x0c;
	break;
    case 0x3f:
	syncinfo.bsid = 0x0d;
	break;
    default:
	syncinfo.bsid = 0x00;
	break;
    }

    syncinfo.frame_size =
	(((uint_16)(data[5] & 0x03) << 12) |
	 ((uint_16)(data[6] & 0xff) <<  4) |
	 ((uint_16)(data[7] & 0xf0) >>  4)) + 1;
}

// This function requires two arguments:
//   first is the start of the data segment
//   second is the tail of the data segment
//   Note: the start will be shift by the amount
//   for a full DTS frame found in the segment.
static frame_t & scan_dts_frame(const uint_8 *&out, const uint_8 *const tail)
{
    static size_t pos = 4;
    static uint_32 syncword = 0xffffffff;
    static syncinfo_t syncinfo;
    static size_t payload_size = 0;

    dts.play.burst = (uint_32 *)0;
    dts.play.size = 0;
resync:
    //
    // Find an DTS sync frame.
    //
    while(syncword != magic) {
	if (out >= tail)
	    goto done;
	syncword = (syncword << 8) | *out++;
    }

    //
    // Need the next 4 bytes to decide how big
    // and which mode the frame is
    //
    while(pos < 8) {
	if (out >= tail)
	    goto done;
	dts.payload[pos++] = *out++;
    }

    //
    // Parse and check syncinfo if not known for this frame
    //
    if (!payload_size) {
	parse_syncinfo(syncinfo, dts.payload);
	if (!syncinfo.frame_size) {
	    syncword = 0xffffffff;
	    pos = 4;
	    esyslog("DTSPCM: ** Invalid frame found - try to syncing **");
	    if (out >= tail)
		goto done;
	    else
		goto resync;
	}
	payload_size = (uint_32)((syncinfo.frame_size));
    }

    //
    // Maybe: Reinit spdif interface if sample rate is changed
    //
    /* ?? Anybody familiar with DTS header format ??? */

    while(pos < payload_size) {
	int rest = tail - out;
	if (rest <= 0)
	    goto done;
	if (rest + pos > payload_size)
	    rest = payload_size - pos;
	memcpy(&dts.payload[pos], out, rest);
	pos += rest;
	out += rest;
    }

    //
    // Check the crc over the entire frame
    //
    /* ?? Anybody familiar with DTS header format ??? */

    //
    // If we reach this point, we found a valid DTS frame to warp into PCM (IEC60958)
    // accordingly to IEC61937
    //
    // What | PCM head              |    DTS      payload
    // pos  | 0,1 | 2,3 | 4,5 | 6,7 | 8,9,10,11 |
    // size |  16 |  16 |  16 |  16 |  16 |  16 |
    // pos  |                       | 0,1 | 2,3 |
    // mean |                       |   sync    |
    //
    dts.pcm[0] = char2short(0xf8, 0x72);			// Pa
    dts.pcm[1] = char2short(0x4e, 0x1f);			// Pb
    dts.pcm[2] = char2short(0x00, syncinfo.bsid);		// DTS data
    dts.pcm[3] = shorts(syncinfo.frame_size * 8);		// Trailing DTS frame size
    dts.pcm[4] = char2short(0x7f, 0xfe);			// DTS first syncwork
    dts.pcm[5] = char2short(0x80, 0x01);			// DTS second syncwork
    swab(dts.payload + 4, dts.payload + 4, payload_size - 4);
    // With padding zeros and do not overwrite IEC958 head
    memset(dts.payload + payload_size, 0,  DTS_BURST_SIZE - 8 - payload_size);

    dts.play.burst = (uint_32 *)(&dts.current[0]);
    dts.play.size = DTS_BURST_SIZE;
    dts.play.pay = payload_size;

    dts.Switch();

    dts.last.burst = (uint_32 *)(&dts.remember[0]);
    dts.last.size = dts.play.size;
    dts.last.pay = dts.play.pay;

    //
    // Reset the syncword for next time
    //
    syncword = 0xffffffff;
    pos = 4;
    payload_size = 0;
done:
    return dts.play;
}
