/*
 * pts.h:	P(resentation) T(ime) S(tamps) handling
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2003 Werner Fink, <werner@suse.de>
 *
 * Changes and Additional Code:
 *
 * Copyright (C) 2003 Sven Goethel, <sven@jausoft.com>
 */

#ifndef _PTS_H
#define _PTS_H

#include <unistd.h>
#include <stdint.h>
#include <sys/time.h>
#include "types.h"
#include "any_decoder.h"

# define debug_pts(args...)
// # define debug_pts(args...)	fprintf(stderr, args)

enum ePTSstate { eOK = 0, eMISSED = -1, eDOSKIP = 1 };

class cPTS {
private:
    // to hold the current state
    uint_32 start, now;
    struct timeval ts;

    // to process and to hold the cummulative state
    uint_32 l_start, l_now;
    long    slept_delay;

public:
    static inline uint_32 PTSinMS(const uchar *pts)
    {
	return (uint_32)
	((   ((((uint64_t)pts[0]) & 0x0E) << 29)
	   | (( (uint64_t)pts[1])         << 22)
	   | ((((uint64_t)pts[2]) & 0xFE) << 14)
	   | (( (uint64_t)pts[3])         <<  7)
	   | ((((uint64_t)pts[4]) & 0xFE) >>  1))/90);
    }

/*

    sender (VDR)                    <->          receiver (spdif)
        |                                            |
     forward credits (ms)            ->           use credits if not expired
   + current time stamp for expire                if no data for credits repeat last frame

                                    <-            if expired put backward credits
                                                  of repeated frames

     if expired get backward credits
     for skipping                    ->           do skipping 
 */

    cPTS(void) 
    : start(0), now(0), 
      l_start(0), l_now(0), 
      slept_delay(1)
    { };

    inline void mark(const uchar *pts)
    {
	start = PTSinMS(pts);
    };
    inline ePTSstate compare(const uchar *pts, const uint_32 tolerance)
    {
	mark(pts);
	bool lower = (now - tolerance <= start);
	bool upper = (start <= now + tolerance);
	ePTSstate ret;

	if (upper && lower)
	    ret = eOK;			// OK
	else if (!upper)
	    ret = eMISSED;		// MISSED frames
	else //  !lower
	    ret = eDOSKIP;		// DOSKIP frames
	return ret;
    };
    inline void clear()
    {
	debug_pts("PTS data cleared\n");
	start = 0;
	l_start=0;
	l_now  =0;
        slept_delay=1;
	reset();
    }
    inline void reset()
    {
	debug_pts("PTS now <- start\n");
	now = start;
	gettimeofday(&ts, 0);
    };
    inline void done()
    {
	struct timeval cur, diff;
	gettimeofday(&cur, 0);
	timersub(&cur, &ts, &diff);
	now += (diff.tv_sec*1000+diff.tv_usec/1000);
	gettimeofday(&ts, 0);
    };
    inline uint_32 delay()
    {
	return now - start;
    };
    inline uint_32 getNow() { return now; }

    inline uint_32 getStart() { return start; }

    
    inline void process( const uchar *pts, AnyDecoder * AnyDecOptional )
    {
	long    delay_pts=0, delay_intern=0;

	if ( pts==NULL) 
		goto out;

	// if cleared -> setup
	if ( start==0 ) {
		debug_pts("PTS mark reset\n");
		mark(pts); // start pts setup
		reset();	    // start -> now
	} else
		done(); // notify time now: how much has been seen !

	mark(pts);

	if(l_start==0 || l_now==0) {
		l_start=start;
		l_now  =now;
	}

	if ( AnyDecOptional != NULL )
		AnyDecOptional->setDelay(0);

	debug_pts("PTS dn %4lums, ds %4lums, ",
		now-l_now, start-l_start);

	if ( now > start ) {
		debug_pts(" late  now>start: %4lums - %4lums = %4lums, ", now, start, now-start);
	} else {
		debug_pts(" early start>now: %4lums - %4lums = %4lums, ", start, now, start-now);
	}

/**
	delay_intern += (long)(now-l_now) - (long)(start-l_start);
		
	if ( now > start ) {
		delay_pts = (long) ( now-start ) + delay_intern;
	} else {
		delay_pts = ((long) ( start-now ) + delay_intern) *-1;
	}
*/

	delay_intern = (long)(now-l_now) - (long)(start-l_start);
	if ( delay_intern > 0 )
		now -= (uint_32) delay_intern;
	else
		now += (uint_32)(delay_intern*-1);
		
	if ( now > start ) {
		delay_pts = (long) ( now-start ) ;
	} else {
		delay_pts = (long) ( start-now ) * -1 ;
	}

	debug_pts("d %4ld/%4ld\n", delay_intern, delay_pts);

/***
	if ( delay_pts > 0 ) {
		debug_pts("PTSDIFF LATE  now>start: %4lums - %4lums = %4lums\n", now, start, now-start);
		if ( AnyDecOptional != NULL )
			AnyDecOptional->setDelay((uint_32)delay_pts);
	}  
	else if ( delay_pts < -slept_delay ) {
		debug_pts("PTSDIFF EARLY start>now: %4lums - %4lums = %4lums, ", start, now, start-now);
		delay_pts*=-1;

		if ( delay_pts > slept_delay)
			usleep ( (uint_32)(delay_pts-slept_delay) * 1000 );
		else
			usleep ( 1000 );
		l_now = now;
		done();
		if( slept_delay == 1 )
			slept_delay = (long)(now-l_now);
		debug_pts(" slept %4lums (%4ld)\n", now-l_now, slept_delay);
	}
 */

      out:

	l_start=start;
	l_now  =now;
    }
};

#endif

